/*************************************************************************

"FreePastry" Peer-to-Peer Application Development Substrate 

Copyright 2002, Rice University. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

- Neither  the name  of Rice  University (RICE) nor  the names  of its
contributors may be  used to endorse or promote  products derived from
this software without specific prior written permission.

This software is provided by RICE and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall RICE or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

********************************************************************************/
package rice.p2p.glacier;

import rice.*;
import rice.p2p.commonapi.*;

/**
 * @(#) VersioningPast.java This interface is exported by a PAST instance that
 * offers access to specific versions of an object.
 *
 * @version $Id: VersioningPast.java,v 1.3 2004/09/03 23:36:52 ahae Exp $
 * @author Andreas Haeberlen
 */

public interface VersioningPast {

  /**
   * Retrieves the object stored in this instance of Past with the given ID and
   * the specified version. Asynchronously returns a PastContent object as the
   * result to the provided Continuation, or a PastException.
   *
   * @param id the key to be queried @
   * @param version the requested version
   * @param command Command to be performed when the result is received
   */
  public void lookup(Id id, long version, Continuation command);

  /**
   * Retrieves the handles of up to max replicas of the object stored in this
   * instance of Past with the given ID. Asynchronously returns an array of
   * PastContentHandles as the result to the provided Continuation, or a
   * PastException. Each replica handle is obtained from a different primary
   * storage root for the the given key. If max exceeds the replication factor r
   * of this Past instance, only r replicas are returned. This method will
   * return a PastContentHandle[] array containing all of the handles.
   *
   * @param id the key to be queried
   * @param version the requested version
   * @param command Command to be performed when the result is received
   * @param num DESCRIBE THE PARAMETER
   */
  public void lookupHandles(Id id, long version, int num, Continuation command);

  /**
   * Updates the objects stored under the provided keys id to expire no earlier
   * than the provided expiration time. Asyncroniously returns the result to the
   * caller via the provided continuation. The result of this operation is an
   * Object[], which is the same length as the input array of Ids. Each element
   * in the array is either Boolean(true), representing that the refresh
   * succeeded for the cooresponding Id, or an Exception describing why the
   * refresh failed. Specifically, the possible exceptions which can be returned
   * are: ObjectNotFoundException - if no object was found under the given key
   * RefreshFailedException - if the refresh operation failed for any other
   * reason (the getMessage() will describe the failure)
   *
   * @param command Command to be performed when the result is received
   * @param ids DESCRIBE THE PARAMETER
   * @param versions DESCRIBE THE PARAMETER
   * @param expirations DESCRIBE THE PARAMETER
   */
  public void refresh(Id[] ids, long[] versions, long[] expirations, Continuation command);

}

