/*************************************************************************

"FreePastry" Peer-to-Peer Application Development Substrate 

Copyright 2002, Rice University. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

- Neither  the name  of Rice  University (RICE) nor  the names  of its
contributors may be  used to endorse or promote  products derived from
this software without specific prior written permission.

This software is provided by RICE and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall RICE or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

********************************************************************************/

package rice.p2p.multiring;

import rice.p2p.commonapi.*;

/**
 * @(#) MultiringApplication.java Application which wraps a real application
 *
 * @version $Id: MultiringApplication.java,v 1.3 2005/03/11 00:58:01 jeffh Exp $
 * @author Alan Mislove
 * @author Peter Druschel
 */
public class MultiringApplication implements Application {

  /**
   * The app which this mulitring app is wrapping
   */
  protected Application application;

  /**
   * The Id which represents the current ring this app is a member of
   */
  protected Id ringId;

  /**
   * Constructor
   *
   * @param ringId DESCRIBE THE PARAMETER
   * @param application DESCRIBE THE PARAMETER
   */
  protected MultiringApplication(Id ringId, Application application) {
    this.application = application;
    this.ringId = ringId;

    if ((ringId instanceof RingId) || (application instanceof MultiringApplication)) {
      throw new IllegalArgumentException("Illegal creation of MRApplication: " + ringId.getClass() + ", " + application.getClass());
    }
  }

  /**
   * This method is invoked on applications when the underlying node is about to
   * forward the given message with the provided target to the specified next
   * hop. Applications can change the contents of the message, specify a
   * different nextHop (through re-routing), or completely terminate the
   * message.
   *
   * @param message The message being sent, containing an internal message along
   *      with a destination key and nodeHandle next hop.
   * @return Whether or not to forward the message further
   */
  public boolean forward(RouteMessage message) {
    return application.forward(new MultiringRouteMessage(ringId, message));
  }

  /**
   * This method is called on the application at the destination node for the
   * given id.
   *
   * @param id The destination id of the message
   * @param message The message being sent
   */
  public void deliver(Id id, Message message) {
    if (id != null) {
      application.deliver(RingId.build(ringId, id), message);
    } else {
      application.deliver(null, message);
    }
  }

  /**
   * This method is invoked to inform the application that the given node has
   * either joined or left the neighbor set of the local node, as the set would
   * be returned by the neighborSet call.
   *
   * @param handle The handle that has joined/left
   * @param joined Whether the node has joined or left
   */
  public void update(NodeHandle handle, boolean joined) {
    application.update(new MultiringNodeHandle(ringId, handle), joined);
  }

}


