/*************************************************************************

"FreePastry" Peer-to-Peer Application Development Substrate 

Copyright 2002, Rice University. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

- Neither  the name  of Rice  University (RICE) nor  the names  of its
contributors may be  used to endorse or promote  products derived from
this software without specific prior written permission.

This software is provided by RICE and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall RICE or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

********************************************************************************/

package rice.p2p.past;

import java.io.Serializable;

import rice.*;
import rice.p2p.commonapi.*;

/**
 * @(#) PastContentHandle.java This interface must be implemented by all content
 * object handles. This interface represents an object that is stored on a
 * specific replica. Thus, the validity of a handle is sometimes very short, and
 * applications should keep references to an object by Id rather than handle.
 *
 * @version $Id: PastContentHandle.java,v 1.5 2005/03/11 00:58:03 jeffh Exp $
 * @author Alan Mislove
 * @author Ansley Post
 * @author Peter Druschel
 */
public interface PastContentHandle extends Serializable {

  /**
   * get the id of the PastContent object associated with this handle
   *
   * @return the id
   */
  public Id getId();

  /**
   * get the NodeHandle of the Past node on which the object associated with
   * this handle is stored
   *
   * @return the id
   */
  public NodeHandle getNodeHandle();

}



